@@ -2,7 +2,7 @@
    +----------------------------------------------------------------------+
    | Xdebug                                                               |
    +----------------------------------------------------------------------+
-   | Copyright (c) 2002-2022 Derick Rethans                               |
+   | Copyright (c) 2002-2023 Derick Rethans                               |
    +----------------------------------------------------------------------+
    | This source file is subject to version 1.01 of the Xdebug license,   |
    | that is bundled with this package in the file LICENSE, and is        |
@@ -19,6 +19,7 @@
 #include "php_xdebug.h"
 #include "tracing_private.h"
 #include "trace_textual.h"
+#include "trace_flamegraph.h"
 #include "trace_computerized.h"
 #include "trace_html.h"
 
@@ -37,11 +38,20 @@ static xdebug_trace_handler_t *xdebug_select_trace_handler(int options)
 		case 0: tmp = &xdebug_trace_handler_textual; break;
 		case 1: tmp = &xdebug_trace_handler_computerized; break;
 		case 2: tmp = &xdebug_trace_handler_html; break;
+		case 3: tmp = &xdebug_trace_handler_flamegraph_cost; break;
+		case 4: tmp = &xdebug_trace_handler_flamegraph_mem; break;
 		default:
 			php_error(E_NOTICE, "A wrong value for xdebug.trace_format was selected (%d), defaulting to the textual format", (int) XINI_TRACE(trace_format));
 			tmp = &xdebug_trace_handler_textual; break;
 	}
 
+	/* Override handler based on options */
+	if (options & XDEBUG_TRACE_OPTION_FLAMEGRAPH_COST) {
+		tmp = &xdebug_trace_handler_flamegraph_cost;
+	}
+	if (options & XDEBUG_TRACE_OPTION_FLAMEGRAPH_MEM) {
+		tmp = &xdebug_trace_handler_flamegraph_mem;
+	}
 	if (options & XDEBUG_TRACE_OPTION_COMPUTERIZED) {
 		tmp = &xdebug_trace_handler_computerized;
 	}
@@ -655,6 +665,8 @@ void xdebug_tracing_register_constants(INIT_FUNC_ARGS)
 {
 	REGISTER_LONG_CONSTANT("XDEBUG_TRACE_APPEND", XDEBUG_TRACE_OPTION_APPEND, CONST_CS | CONST_PERSISTENT);
 	REGISTER_LONG_CONSTANT("XDEBUG_TRACE_COMPUTERIZED", XDEBUG_TRACE_OPTION_COMPUTERIZED, CONST_CS | CONST_PERSISTENT);
+	REGISTER_LONG_CONSTANT("XDEBUG_TRACE_FLAMEGRAPH_COST", XDEBUG_TRACE_OPTION_FLAMEGRAPH_COST, CONST_CS | CONST_PERSISTENT);
+	REGISTER_LONG_CONSTANT("XDEBUG_TRACE_FLAMEGRAPH_MEM", XDEBUG_TRACE_OPTION_FLAMEGRAPH_MEM, CONST_CS | CONST_PERSISTENT);
 	REGISTER_LONG_CONSTANT("XDEBUG_TRACE_HTML", XDEBUG_TRACE_OPTION_HTML, CONST_CS | CONST_PERSISTENT);
 	REGISTER_LONG_CONSTANT("XDEBUG_TRACE_NAKED_FILENAME", XDEBUG_TRACE_OPTION_NAKED_FILENAME, CONST_CS | CONST_PERSISTENT);
 }
@@ -663,6 +675,8 @@ void xdebug_tracing_rinit(void)
 {
 	XG_TRACE(trace_handler) = NULL;
 	XG_TRACE(trace_context) = NULL;
+
+	xdebug_disable_opcache_optimizer();
 }
 
 void xdebug_tracing_post_deactivate(void)
@@ -683,18 +697,18 @@ void xdebug_tracing_init_if_requested(zend_op_array *op_array)
 	}
 }
 
-void xdebug_tracing_execute_ex(int function_nr, function_stack_entry *fse)
+void xdebug_tracing_execute_ex(function_stack_entry *fse)
 {
 	if (fse->filtered_tracing || !XG_TRACE(trace_context)) {
 		return;
 	}
 
 	if (XG_TRACE(trace_handler)->function_entry) {
-		XG_TRACE(trace_handler)->function_entry(XG_TRACE(trace_context), fse, function_nr);
+		XG_TRACE(trace_handler)->function_entry(XG_TRACE(trace_context), fse);
 	}
 }
 
-void xdebug_tracing_execute_ex_end(int function_nr, function_stack_entry *fse, zend_execute_data *execute_data)
+void xdebug_tracing_execute_ex_end(function_stack_entry *fse, zend_execute_data *execute_data, zval *return_value)
 {
 	zend_op_array *op_array;
 
@@ -703,7 +717,7 @@ void xdebug_tracing_execute_ex_end(int function_nr, function_stack_entry *fse, z
 	}
 
 	if ((XG_TRACE(trace_handler)->function_exit)) {
-		XG_TRACE(trace_handler)->function_exit(XG_TRACE(trace_context), fse, function_nr);
+		XG_TRACE(trace_handler)->function_exit(XG_TRACE(trace_context), fse);
 	}
 
 	/* Store return value in the trace file */
@@ -719,42 +733,42 @@ void xdebug_tracing_execute_ex_end(int function_nr, function_stack_entry *fse, z
 
 	if (op_array->fn_flags & ZEND_ACC_GENERATOR) {
 		if (XG_TRACE(trace_handler)->generator_return_value) {
-			XG_TRACE(trace_handler)->generator_return_value(XG_TRACE(trace_context), fse, function_nr, (zend_generator*) execute_data->return_value);
+			XG_TRACE(trace_handler)->generator_return_value(XG_TRACE(trace_context), fse, (zend_generator*) execute_data->return_value);
 		}
 	} else {
 		if (XG_TRACE(trace_handler)->return_value) {
-			XG_TRACE(trace_handler)->return_value(XG_TRACE(trace_context), fse, function_nr, execute_data->return_value);
+			XG_TRACE(trace_handler)->return_value(XG_TRACE(trace_context), fse, return_value);
 		}
 	}
 }
 
-int xdebug_tracing_execute_internal(int function_nr, function_stack_entry *fse)
+int xdebug_tracing_execute_internal(function_stack_entry *fse)
 {
 	if (fse->filtered_tracing || !XG_TRACE(trace_context)) {
 		return 0;
 	}
 
 	if (fse->function.type != XFUNC_ZEND_PASS && (XG_TRACE(trace_handler)->function_entry)) {
-		XG_TRACE(trace_handler)->function_entry(XG_TRACE(trace_context), fse, function_nr);
+		XG_TRACE(trace_handler)->function_entry(XG_TRACE(trace_context), fse);
 		return 1;
 	}
 
 	return 0;
 }
 
-void xdebug_tracing_execute_internal_end(int function_nr, function_stack_entry *fse, zval *return_value)
+void xdebug_tracing_execute_internal_end(function_stack_entry *fse, zval *return_value)
 {
 	if (fse->filtered_tracing || !XG_TRACE(trace_context)) {
 		return;
 	}
 
 	if (fse->function.type != XFUNC_ZEND_PASS && (XG_TRACE(trace_handler)->function_exit)) {
-		XG_TRACE(trace_handler)->function_exit(XG_TRACE(trace_context), fse, function_nr);
+		XG_TRACE(trace_handler)->function_exit(XG_TRACE(trace_context), fse);
 	}
 
 	/* Store return value in the trace file */
 	if (XINI_TRACE(collect_return) && fse->function.type != XFUNC_ZEND_PASS && return_value && XG_TRACE(trace_handler)->return_value) {
-		XG_TRACE(trace_handler)->return_value(XG_TRACE(trace_context), fse, function_nr, return_value);
+		XG_TRACE(trace_handler)->return_value(XG_TRACE(trace_context), fse, return_value);
 	}
 }
 
